8from libtuer import ThreadFunction, logger, fire_and_forget
from actor import Actor
import os, random

# logger.{debug,info,warning,error,critical}

def play_sound (what):
	try:
		soundfiles = os.listdir(SOUNDS_DIRECTORY+what)
	except FileNotFoundError:
		logger.error("StateMachine: Unable to list sound files in %s" % (SOUNDS_DIRECTORY+what))
		return
	soundfile = SOUNDS_DIRECTORY + what + '/' + random.choice(soundfiles)
	fire_and_forget ([SOUNDS_PLAYER,soundfile], logger.error, "StateMachine: ")


# StateUnlocking constants
OPEN_REPEAT_TIMEOUT = 8
OPEN_REPEAT_NUMBER = 3

# StateLocking constants
CLOSE_REPEAT_TIMEOUT = 8
CLOSE_REPEAT_NUMBER = 3

# StateAboutToOpen constants
ABOUTOPEN_NERVLIST = [(5, lambda : play_sound("flipswitch")), (5, lambda:play_sound("flipswitch")), (0, lambda:Logger.warning("Space open but switch not flipped for 10 seconds")),\
	(10, lambda:play_sound("flipswitch")), (10, lambda:play_sound("flipswitch")), (0, lambda:Logger.error("Space open but switch not flipped for 30 seconds")),\
	(10, lambda:play_sound("flipswitch")), (10, lambda:play_sound("flipswitch")), (6, lambda:play_sound("flipswitch")), (4, lambda:Logger.critical("Space open but switch not flipped for 60 seconds"))]

# StateAuf constants
#  time that must pass between two bell_ringing events to buzz the door again (seconds)
AUF_BELLBUZZ_REPEAT_TIME = 2

# Timeout we wait after the switch was switched to "Closed", until we assume nobody will open the door and we just lock it
LEAVE_TIMEOUT = 30

# play_sound constants
SOUNDS_DIRECTORY = "/opt/tuer/sounds/"
SOUNDS_PLAYER = "/usr/bin/mplayer"


class Nerver():
	# A little utility class used to run the nervlists. A nervlist is a list of (n, f) tuples where f() is run after n seconds.
	# If f returns something, that's also returned by nerv.
	def __init__(self, nervlist):
		self.nervlist = list(nervlist)
		self.last_event_time = time.time()
	
	def nerv(self):
		if len(self.nervlist):
			(time, f) = self.nervlist[0]
			now = time.time()
			time_gone = now-self.last_event_time
			# check if the first element is to be triggered
			if time_gone >= time:
				self.nervlist = self.nervlist[1:] # "pop" the first element, but do not modify original list
				self.last_event_time = now
				return f()


class StateMachine():
	# commands you can send
	CMD_PINS = 0
	CMD_BUZZ = 1
	CMD_UNLOCK = 2
	CMD_WAKEUP = 3
	CMD_LAST = 4
	
	class State():
		def __init__(self, state_machine, nervlist = None):
			self.state_machine = state_machine
			self._nerver = None if nervlist is None else Nerver(nervlist)
		def handle_pins_event(self):
			pass # one needn't implement this
		def handle_buzz_event(self,arg): # this shouldn't be overwritten
			self.actor.act(Actor.CMD_BUZZ)
			arg("200 okay: buzz executed")
		def handle_cmd_unlock_event(self,arg):
			if arg is not None:
				arg("412 Precondition Failed: The current state (%s) cannot handle the OPEN event" % self.__class__.__name__)
		def handle_wakeup_event(self):
			if self._nerver is not None:
				return self._nerver.nerv()
		def pins(self):
			return self.state_machine.pins
		def actor(self):
			return self.state_machine.actor
		def handle_event(self,ev,arg): # don't override
			if arg is CMD_PINS:
				self.handle_pins_event()
			elif arg is CMD_BUZZ:
				self.handle_buzz_event(arg)
			elif arg is CMD_UNLOCK:
				self.handle_cmd_unlock_event(arg)
			elif arg is CMD_WAKEUP:
				self.handle_wakeup_event()
			else:
				raise Exception("Unknown command number: %d" % ev)
	
	class StateStart(State):
		def handle_pins_event(self):
			super().handle_pins_event()
			thepins = self.pins()
			for pin in thepins:
				if pin is None:
					return None
			if thepins.door_locked:
				return StateZu
			else:
				return StateAuf

	class StateZu(State):
		def handle_pins_event(self):
			super().handle_pins_event()
			pins = self.pins()
			if not pins.door_locked:
				return StateAboutToOpen(self.state_machine)
		def handle_cmd_unlock_event(self,callback):
			# intentionally not calling super() implementation
			return StateUnlocking(callback,self.state_machine)
	
	class StateUnlocking(State):
		def __init__(self,callback,sm):
			# construct a nervlist
			nervlist = [(OPEN_REPEAT_TIMEOUT, lambda: self.actor().act(Actor.CMD_UNLOCK)) for t in xrange(OPEN_REPEAT_NUMBER)]
			nervlist += [(OPEN_REPEAT_TIMEOUT, self.could_not_open)]
			super().__init__(sm,nervlist)
			self.callbacks=[callback]
			# FIXME: can we send "202 processing: Trying to open the door" here? Are the callbacks multi-use?
			self.actor().act(Actor.CMD_UNLOCK)
		def notify(self, did_it_work):
			s = "200 okay: door open" if did_it_work else ("500 internal server error: Couldn't open door with %d tries à %f seconds" % (OPEN_REPEAT_NUMBER,OPEN_REPEAT_TIMEOUT))
			for cb in self.callbacks:
				if cb is not None: # FIXME why this check? shouldn't be needed
					cb(s)
		def handle_pins_event(self):
			super().handle_pins_event()
			pins = self.pins()
			if not pins.door_locked:
				self.notify(True)
				return StateAboutToOpen(self.state_machine)
		def handle_cmd_unlock_event(self,callback):
			# intentionally not calling super() implementation
			# FIXME: 202 notification also here if possible
			self.callbacks.append(callback)
		def could_not_open(self):
			logger.critical("Couldn't open door after %d tries. Going back to StateZu." % OPEN_REPEAT_NUMBER)
			self.notify(False)
			return StateZu(self.state_machine)
	
	class AbstractStateWhereOpeningIsRedundant(State):
		def handle_cmd_unlock_event(self, callback):
			# intentionally not calling super() implementation
			callback("299 redundant: Space seems to be already open. Still processing your request tough.")
			logger.warning("Received OPEN command in StateAboutToOpen. This should not be necessary.")
			self.actor().act(Actor.CMD_UNLOCK)
	
	class StateAboutToOpen(AbstractStateWhereOpeningIsRedundant):
		def __init__(self, sm):
			super().__init__(sm, ABOUTOPEN_NERVLIST)
		def handle_pins_event(self):
			super().handle_pins_event()
			pins = self.pins()
			if pins.door_locked:
				return StateZu(self.state_machine)
			elif pins.space_active:
				return StateAuf(self.state_machine)
	
	class StateAuf(AbstractStateWhereOpeningIsRedundant):
		def __init__(self,sm):
			super().__init__(sm)
			self.last_buzzed = None
		def handle_pins_event(self):
			super().handle_pins_event()
			pins = self.pins()
			if pins.bell_ringing:
				now = time.time()
				if self.last_buzzed is None or now-self.last_buzzed < AUF_BELLBUZZ_REPEAT_TIME:
					logger.info("buzzing because of bell ringing in state auf")
					self.actor().act(Actor.CMD_BUZZ)
					self.last_buzzed = now
			if not pins.space_active:
				logger.info("space switch off - starting leaving procedure")
				return StateAboutToLeave(self.state_machine)
			if pins.door_locked:
				logger.error("door manually locked, but space switch on - going to StateZu")
				play_sound("manual_lock")
				return StateZu(self.state_machine)
	
	class StateLocking(State):
		# FIXME: Why does this even have callbacks?
		# TODO: share code with StateUnlocking
		def __init__(self,sm):
			# construct a nervlist
			nervlist = [(t*CLOSE_REPEAT_TIMEOUT, lambda: self.actor().act(Actor.CMD_LOCK)) for t in range(1, CLOSE_REPEAT_NUMBER+1)]
			nervlist += [((CLOSE_REPEAT_NUMBER+1)*CLOSE_REPEAT_TIMEOUT, self.could_not_close)]
			super().__init__(sm, nervlist)
			self.callbacks=[]
			# FIXME: can we send "202 processing: Trying to close the door" here? Are the callbacks multi-use?
			self.tries = 0
			assert self.pins().door_closed, "Door is open while we should close it, this must not happen"
			self.actor().act(Actor.CMD_LOCK)
		def notify(self, did_it_work):
			s = "200 okay: door closed" if did_it_work else ("500 internal server error: Couldn't close door with %d tries à %f seconds" % (CLOSE_REPEAT_NUMBER,CLOSE_REPEAT_TIMEOUT))
			for cb in self.callbacks:
				if cb is not None:
					cb(s)
		def handle_pins_event(self):
			pins = self.pins()
			if not pins.door_closed:
				# TODO play a sound? This shouldn't happen, door was opened while we are locking
				self.notify(True)
				return StateAboutToOpen(self.state_machine)
			if pins.door_locked:
				return SpaceZu(self.state_machine)
		def handle_cmd_unlock_event(self,callback):
			callback("409 conflict: The server is currently trying to close the door. Try again later.")
		def could_not_close(self):
			logger.critical("Couldn't close door after %d tries. Going back to StateAboutToOpen." % CLOSE_REPEAT_NUMBER)
			self.notify(False)
			return StateAboutToOpen(self.state_machine)
	
	class StateAboutToLeave(State):
		def __init__(self, sm):
			nervlist = [(LEAVE_TIMEOUT, lambda: StateLocking(self.state_machine))]
			super().__init__(sm, nervlist)
		def handle_pins_event(self):
			if not self.pins().door_closed:
				return StateLeaving(self.state_machine)
			if self.pins().door_locked:
				return StateZu(self.state_machine)
	
	class StateLeaving(State):
		def __init__(self, sm):
			nervlist = [(LEAVE_TIMEOUT, lambda: StateAboutToOpen(self.state_machine))]
			super().__init__(sm, nervlist)
		def handle_pins_event(self):
			if self.pins().door_closed:
				return StateLocking(self.state_machine)
			if self.pins().door_locked:
				return StateZu(self.state_machine)
	
	def __init__(self, actor):
		self.actor = actor
		self.callback = ThreadFunction(self._callback)
		self.current_state = StateStart(self)
		self.pins = None
		self.old_pins = None
	
	def stop (self):
		self.callback.stop()
	
	def _callback(self, cmd, arg=None):
		# update pins
		if cmd == StateMachine.CMD_PINS:
			self.pins = arg
		# handle stuff
		newstate = self.current_state.handle_event(cmd,arg) # returns None or an instance of the new state
		self.old_pins = self.pins # FIXME not used?
		while newstate is not None:
			logger.info("StateMachine: new state = %s" % newstate.__class__.__name__)
			self.current_state = newstate
			newstate = self.current_state.handle_event(StateMachine.CMD_PINS, self.pins)
